# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Lance Authors

import logging
import os
from typing import List, Optional, TypeVar, Any, Dict

# Importing classes from a hypothetical LAS REST library
from las_rest.core.catalog_client import CatalogClient
from las_rest.core.entity.database_builder import DatabaseBuilder
from las_rest.core.entity.table import FieldColumn, PartitionType, FileFormat
from las_rest.core.entity.table import Table
from las_rest.core.entity.table_builder import TableBuilder

from .catalog import Catalog
from .utils import TableIdentifier
from .utils import Utils

# Setting up the logging configuration
T = TypeVar('T')

# 配置基本的日志格式和级别
logging.basicConfig(
    level=logging.ERROR,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)

# 获取 logger
logger = logging.getLogger(__name__)

# Various constants used for setting up the catalog and managing tables
LIST_ALL_TABLES = "list-all-tables"
LIST_ALL_TABLES_DEFAULT = "false"
TABLE_TYPE_PROP = "table_type"
LANCE_DATASET_TYPE_VALUE = "lance"
TABLE_STORAGE_OPTIONS_PROP = "table_storage"
HMS_TABLE_OWNER = "hive.metastore.table.owner"
HMS_DB_OWNER_TYPE = "hive.metastore.database.owner-type"
HMS_DB_OWNER = "hive.metastore.database.owner"
STORAGE_OPTIONS_AK = "access_key_id"
STORAGE_OPTIONS_SK = "secret_access_key"
STORAGE_OPTIONS_ENDPOINT = "aws_endpoint"
STORAGE_OPTIONS_VIRTUAL_HOSTED = "virtual_hosted_style_request"


class LasCatalog(Catalog):
    """
    A class to manage the LAS Catalog, including creating namespaces, listing catalogs,
    and integrating with AWS services via CatalogClient.
    """

    def __init__(self, access_key=None, secret_key=None, sts_token=None, region=None, catalog_name=None, env="online",
                 storage_config: dict = None):
        """
        Initializes the LasCatalog object either from individual parameters or a dictionary storing all configurations.

        Parameters:
            access_key (str): Access key for authentication -- optional if provided within storage_config.
            secret_key (str): Secret key for authentication -- optional if provided within storage_config.
            sts_token (str): Optional STS token for temporary access.
            region (str): The region where the catalog is hosted.
            catalog_name (str): The name of the catalog.
            env (str): Environment setting, default is "online".
            storage_config (dict): Dictionary containing all necessary configurations.
        """

        def get_param(config_dict, keys, default=None):
            if config_dict:
                for key in [k for k in keys if k in config_dict]:
                    return config_dict[key]
            return default

        # Initialize configuration parameters
        self.access_key = get_param(storage_config, ['access_key', 'access_key_id'], access_key)
        self.secret_key = get_param(storage_config, ['secret_key', 'secret_access_key'], secret_key)
        self.region = get_param(storage_config, ['region', 'aws_region'], region)

        # Directly initialized parameters
        self.sts_token = get_param(storage_config, ['sts_token'], sts_token)
        self.catalog_name = get_param(storage_config, ['catalog_name'], catalog_name)
        self.env = get_param(storage_config, ['env'], env)
        self.catalog_properties = {}

        if not all([self.access_key, self.secret_key, self.region, self.catalog_name]):
            raise ValueError("Missing required parameters: access_key/access_key_id, secret_key/secret_access_key, region/aws_region, catalog_name")


        # Initialize environment variables and CatalogClient for interacting with the LAS catalog
        os.environ['CATALOG_ENVIRONMENT'] = self.env
        self.las_catalog_client = CatalogClient(
            self.access_key,
            self.secret_key,
            self.sts_token,
            self.region,
            self.catalog_name
        )

    @classmethod
    def from_config(cls, config: dict):
        """
        Class method to create an instance of LasCatalog from a configuration dictionary directly.

        Parameters:
            config (dict): Dictionary containing all configurations for the LasCatalog.

        Returns:
            LasCatalog: An instance of LasCatalog.
        """
        return cls(storage_config=config)

    def set_catalog(self, catalog_name: str):
        """
        Sets the catalog name and updates the underlying catalog client.

        Parameters:
            catalog_name (str): Name of the catalog to set.
        """
        self.catalog_name = catalog_name
        self.las_catalog_client.catalog_name = catalog_name  # Assuming CatalogClient has an attribute `catalog_name`


    def list_catalogs(self):
        """
        Retrieves a list of all catalogs available in the LAS  environment.

        This method uses the `las_catalog_client` to make the API call and returns the list of catalogs.

        Returns:
            list: A list of catalog names.

        Example:
            >>> # Assuming 'las_catalog' is an instance of LasCatalog
            >>> catalogs = self.las_catalog_client.list_catalogs()
            >>> print(catalogs)
            ['catalog1', 'catalog2', 'catalog3']

        This method is useful for getting an overview of all the catalogs that are available for use within the LAS system.
        It can be used to programmatically access and manage different catalogs based on the application's requirements.
        """
        catalog_list = self.las_catalog_client.list_catalogs()
        return catalog_list


    def list_databases(self, catalog_name: str):
        """
        List all databases in a given catalog.

        This method takes a catalog name and returns a list of all databases in that catalog.

        Args:
            catalog_name (str): The name of the catalog.

        Returns:
            list: A list of database names.

        Raises:
            Exception: If an error occurs while retrieving the databases.
        """
        try:
            # Try to get all databases from the LAS catalog client
            database_names = self.las_catalog_client.get_all_databases(catalog_name)
            # Return the list of database names
            return database_names
        except Exception as e:
            # Log an error message if the catalog does not exist
            logger.error(f"catalog does not exist: {e}")


    def list_tables(self, database: str) -> list[str]:
        """
        List all tables in a given database.

        This method takes a database name and returns a list of all tables in that database.

        Args:
            database (str): The name of the database.

        Returns:
            list: A list of dataset names.

        Raises:
            Exception: If an error occurs while retrieving the tables.
        """
        try:
            # Try to get all tables from the LAS catalog client
            tbl_names = self.las_catalog_client.get_all_tables(database, self.catalog_name)
            # Return the list of table names
            return tbl_names
        except Exception as e:
            # Log an error message if the database does not exist
            logger.error(f"Database does not exist: {e}")
            return []

    def create_database(self, database_name, location):
        database = DatabaseBuilder(
            name=database_name,
            location=location,
            catalog_name=self.catalog_name
        ).with_description("lance database").build()

        try:
            self.las_catalog_client.create_database(database)
        except Exception as e:
            logger.error(f"Database {database_name} already exists")
            raise e
        return database_name


    def drop_database(self, database_name: str) -> None:
        """Drop a namespace.

        Args:
            database_name: Namespace identifier.

        Raises:
            NoSuchNamespaceError: If a namespace with the given name does not exist, or the identifier is invalid.
            NamespaceNotEmptyError: If the namespace is not empty.
        """
        try:
            self.las_catalog_client.drop_database(database_name, self.catalog_name)
        except Exception as e:
            raise Exception(f"Database {database_name} is not empty") from e


    def drop_table(self, identifier: TableIdentifier, purge=False):
        db_name, tbl_name = Utils.identifier_to_database_and_table(identifier)
        try:
            self.las_catalog_client.drop_table(self.catalog_name, db_name , tbl_name)
            if purge:
                self.purge_data(identifier)

            logger.info(f"Dropped table: {identifier}")
            return True

        except Exception as e:
            logger.info(f"Skipping drop, table does not exist: {identifier} - {str(e)}")
            return False


    def purge_data(self, identifier: TableIdentifier):
        # Example function to delete data given a location
        # Implementation would depend on your data storage details
        pass

    def list_namespaces(self):
        """List namespaces from the given namespace. If not given, list top-level namespaces from the catalog.

        Returns:
            List[Identifier]: a List of namespace identifiers.
        """
        # Hierarchical namespace is not supported. Return an empty list
        return self.las_catalog_client.get_all_databases()


    def get_table(self, identifier: TableIdentifier) -> Optional[Table]:
        db_name, tbl_name = Utils.identifier_to_database_and_table(identifier)
        try:
            return self.las_catalog_client.get_table(tbl_name, db_name, self.catalog_name)
        except Exception as e:
            logger.debug(f"Table not found {db_name}.{tbl_name}", exc_info=e)
            return None


    def table_exists(self, identifier: TableIdentifier) -> bool:
        """
        Check if a table exists in the LAS catalog.

        This method takes a table identifier and returns True if the table exists, False otherwise.

        Args:
            identifier (TableIdentifier): The identifier for the table, which can be a string or a tuple of strings.

        Returns:
            bool: True if the table exists, False otherwise.
        """
        # Convert the identifier to database and table names
        db_name, tbl_name = Utils.identifier_to_database_and_table(identifier)
        try:
            # Try to get the table from the LAS catalog client
            table = self.las_catalog_client.get_table(tbl_name, db_name, self.catalog_name)
            # If the table is not None, return True
            if table is not None:
                return True
            # If the table is None, return False
            else:
                return False
        except Exception as e:
            # Log a debug message if the table is not found
            logger.debug(f"Table not found {db_name}.{tbl_name}", exc_info=e)
            # Return False if an exception occurs
            return False


    def convert_table_to_dataset_uri(self, identifier: TableIdentifier) -> str:
        """
        Convert a table identifier to a dataset URI.

        This method takes a table identifier and returns the location of the table as a dataset URI.

        Args:
            identifier (TableIdentifier): The identifier for the table, which can be a string or a tuple of strings.

        Returns:
            str: The location of the table as a dataset URI.

        Raises:
            RuntimeError: If the table is not found.
        """
        try:
            # Load the table using the provided identifier
            table = self.get_table(identifier)
            # Return the location of the table as the dataset URI
            return table.location
        except Exception as e:
            # Log an error message if the table is not found
            logger.error(f"Table not found", exc_info=e)
            # Raise a RuntimeError with the original exception
            raise RuntimeError(e)


    def register_lance_table(self, identifier: TableIdentifier, field_columns=None):
        """
        Register a LAS catalog table under a specific database.

        This method takes an identifier and optional field columns to register a table in the LAS catalog.
        The table will be located under the specified database.

        Args:
            identifier (TableIdentifier): The identifier for the table, which can be a string or a tuple of strings.
            field_columns (List[FieldColumn], optional): A list of FieldColumn objects defining the table schema. Defaults to None.

        Returns:
            Table: The newly registered table.

        Raises:
            NoSuchTableError: If the identifier is invalid or does not exist.
            TableAlreadyExistsError: If the table already exists.
        """
        # Convert the identifier to database and table names
        db_name, tbl_name = Utils.identifier_to_database_and_table(identifier)

        # Get the database object from the LAS catalog client
        db = self.las_catalog_client.get_database(db_name, self.catalog_name)

        # Construct the table location by appending the table name to the database location
        tbl_location = db.location + "/" + tbl_name + ".lance"

        # Register the table with the constructed location and field columns
        return self.register_lance_table_location(identifier, tbl_location, field_columns)



    def register_lance_table_location(self, identifier: TableIdentifier, tbl_location, field_columns=None):
        """
        Register a new table in the LAS catalog.

        This method takes an identifier, a table location, and optional field columns to create a new table in the LAS catalog.
        If field columns are not provided, a default schema with a single column "_rowid" of type "bigint" is used.

        Args:
            identifier (TableIdentifier): The identifier for the table, which can be a string or a tuple of strings.
            tbl_location (str): The location where the table data will be stored.
            field_columns (List[FieldColumn], optional): A list of FieldColumn objects defining the table schema. Defaults to None.

        Raises:
            NoSuchTableError: If the identifier is invalid or does not exist.
            Exception: If there is an error creating the table in the Hive Metastore.

        Returns:
            None
        """
        db_name, tbl_name = Utils.identifier_to_database_and_table(identifier)
        if field_columns is None:
            field_columns = [
                FieldColumn("_rowid", "bigint", "default lance schema when no schema is provided")
            ]
        partition_columns = []
        tbl_property = {TABLE_TYPE_PROP:LANCE_DATASET_TYPE_VALUE}
        try:
            tbl = (TableBuilder()
                   .with_catalog_name(self.catalog_name)  # Required
                   .with_database_name(db_name)  # Required
                   .with_table_name(tbl_name)  # Required
                   .with_partition_type(PartitionType.NON_PARTITION)
                   .with_field_columns(field_columns)  # Required
                   .with_partition_columns(partition_columns)
                   .with_file_format(FileFormat.LANCE)  # Required
                   .with_table_properties(tbl_property)
                   .with_table_comment("Lance Catalog Table")
                   .with_location(tbl_location)
                   .build())

            self.las_catalog_client.create_table(tbl)
        except Exception as e:
            logging.error(f"Failed to create table {tbl_name} in Hive Metastore: {str(e)}")
            raise
